/*
 * A pipe pager providing access to pipe content and navigation support.
 *
 * Copyright (C) 2021, 2022, 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "pipe_pager.h"
#include "pipe_object_server.h"



/* Initialise a pager for a pipe with a shared page mapper for moderating
   access to loaded pages. At first, no mapper will be configured: this must be
   done by requesting a region. */

PipePager::PipePager(PipePaging *paging, bool writing)
: Pager(NULL, writing ? L4RE_DS_MAP_FLAG_RW : L4RE_DS_MAP_FLAG_RO),
  _paging(paging), _writing(writing)
{
    /* Initialise the size of the paged region. */

    _size = _paging->region_size();

    /* Initialise any recorded position in the mapped region for this
       endpoint. */

    _data_current = 0;

    /* Obtain any initial page mapper, this having been set up in the paging
       coordinator. */

    _mapper = _paging->current_region(_writing);
}

ipc_server_default_config_type PipePager::config()
{
    return config_PipeObject;
}



/* Close the pager, detaching from the paging coordinator for the pipe. When
   both pagers have detached, this will release all active page mappers and
   related resources. */

void PipePager::close()
{
    if (_paging == NULL)
        return;

    /* Notify the other endpoint and unsubscribe. */

    _paging->notify_others(_writing ? PipePaging::WRITER : PipePaging::READER,
                           NOTIFY_PEER_CLOSED, NOTIFY_VALUES_NULL);
    unsubscribe();

    /* Deallocate the paging coordinator if no other endpoints are active. */

    if (!_paging->detach())
    {
        delete _paging;
        _paging = NULL;
    }
}

/* Support paging. */

long PipePager::map(offset_t offset, map_address_t hot_spot, map_flags_t flags, l4_snd_fpage_t *region)
{
    return Pager::map(offset, hot_spot, flags, region);
}



/* Return whether the pipe is closed or partly closed. */

long PipePager::closed(int *closed)
{
    *closed = _paging->closed();
    return L4_EOK;
}

/* Return details of the current region. */

long PipePager::refresh(offset_t *position, offset_t *start_pos,
                        offset_t *end_pos, offset_t *size,
                        object_flags_t *object_flags)
{
    if (_mapper != NULL)
    {
        if (position != NULL)
            *position = _data_current;

        if (start_pos != NULL)
            *start_pos = 0;

        *end_pos = _size;
        *size = _mapper->get_data_size();

        if (object_flags != NULL)
            *object_flags = 0;

        return L4_EOK;
    }
    else
        return pipe_error();
}

/* Obtain the next region and its details. */

long PipePager::next_region(offset_t *populated_size, offset_t *size)
{
    /* Obtain a new region if writing. */

    if (_writing)
        return next_region_for_writer(populated_size, size);
    else
        return next_region_for_reader(populated_size, size);
}

long PipePager::next_region_for_reader(offset_t *populated_size, offset_t *size)
{
    PageMapper *mapper = _paging->next_region();

    if (mapper == NULL)
        return pipe_error();

    _mapper = mapper;
    _data_current = 0;

    return refresh(NULL, NULL, size, populated_size, NULL);
}

long PipePager::next_region_for_writer(offset_t *populated_size, offset_t *size)
{
    /* Set the populated size before moving on. */

    if (_mapper != NULL)
        _mapper->set_data_size(*populated_size);

    /* Obtain the page mapper for the region. */

    PageMapper *mapper = _paging->add_region();

    if (mapper == NULL)
        return pipe_error();

    _mapper = mapper;
    _data_current = 0;

    return refresh(NULL, NULL, size, populated_size, NULL);
}

long PipePager::pipe_error()
{
    if (_paging->closed())
        return -L4_EIO;
    else
        return -L4_EBUSY;
}



/* Update the consumed or populated size of a pipe region and notify the other
   endpoint. The consumed size is recorded to allow pipes to be shared between
   programs. */

long PipePager::flush(offset_t position, offset_t *size)
{
    if (_mapper != NULL)
    {
        _data_current = position;

        if (_writing)
            _mapper->set_data_size(position);
    }

    *size = _size;

    // NOTE: Perhaps employ a distinct event type for metadata updates.

    _paging->notify_others(_writing ? PipePaging::WRITER : PipePaging::READER,
                           NOTIFY_CONTENT_AVAILABLE | NOTIFY_SPACE_AVAILABLE,
                           NOTIFY_VALUES_NULL);
    return L4_EOK;
}



/* Subscribe to notifications.
   Readers can subscribe to new data (at end), and pipe closed events.
   Writers can subscribe to new space and pipe closed events. */

long PipePager::subscribe(l4_cap_idx_t notifier, notify_flags_t flags)
{
    /* A single notifier is recorded so that it may be unsubscribed when the
       endpoint is closed. */

    unsubscribe();

    _notifier = notifier;
    return _paging->subscribe(_writing ? PipePaging::WRITER : PipePaging::READER,
                              _notifier, flags);
}

/* Unsubscribe from notifications. */

long PipePager::unsubscribe()
{
    if (l4_is_valid_cap(_notifier))
    {
        long err = _paging->unsubscribe(_notifier);
        _notifier = L4_INVALID_CAP;
        return err;
    }
    else
        return L4_EOK;
}

// vim: tabstop=4 expandtab shiftwidth=4
